package org.expeditee.actions;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GraphicsEnvironment;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;

import javax.imageio.ImageIO;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.FrameMouseActions;
import org.expeditee.items.Item;
import org.expeditee.items.Text;
import org.expeditee.stats.SessionStats;
import org.expeditee.stats.StatsLogger;

/**
 * A list of miscellaneous KMS Actions and Actions specific to Expeditee
 * 
 */
public class Misc {
	private static final int STATS_FONT_SIZE = 18;

	private static final int FRAME_FILE_FONT_SIZE = 10;

	/**
	 * Causes the system to beep
	 */
	public static void Beep() {
		java.awt.Toolkit.getDefaultToolkit().beep();
	}

	/**
	 * Forces a repaint of the current Frame
	 */
	public static void Display() {
		FrameGraphics.ForceRepaint();
	}

	/**
	 * Restores the current frame to the last saved version currently on the
	 * hard disk
	 */
	public static void Restore() {
		FrameIO.Reload();

		FrameGraphics.DisplayMessage("Restoration complete.");
	}

	/**
	 * Toggles AudienceMode on or off
	 */
	public static void ToggleAudienceMode() {
		FrameGraphics.ToggleAudienceMode();
	}

	/**
	 * Toggles TwinFrames mode on or off
	 */
	public static void ToggleTwinFramesMode() {
		DisplayIO.ToggleTwinFrames();
	}

	/**
	 * Exits the System
	 */
	public static void Exit() {
		/**
		 * TODO: Prompt the user etc.
		 */
		FrameIO.SaveFrame(DisplayIO.getCurrentFrame());
		
		while(FrameIO.DeleteFrameset("messages"))
			;

		StatsLogger.WriteStatsFile();

		System.exit(0);
	}

	/**
	 * If the given Item is a Text Item, then the text of the Item is
	 * interpreted as actions, if not this method does nothing.
	 * 
	 * @param current
	 *            The Item to read the Actions from
	 */
	public static void RunCurrentItem(Item current) {
		if (current instanceof Text) {
			List<String> actions = ((Text) current).getText();

			for (String action : actions) {
				Actions.PerformAction(DisplayIO.getCurrentFrame(), current,
						action);
			}
		}

	}

	/**
	 * Prompts the user to confirm deletion of the current Frame, and deletes if
	 * the user chooses. After deletion this action calls back(), to ensure the
	 * deleted frame is not still being shown
	 * 
	 */
	public static void DeleteFrame() {
		Frame toDelete = DisplayIO.getCurrentFrame();
		DisplayIO.Back();
		String deletedFrame = toDelete.getFrameName();
		String deletedFrameNameLowercase = deletedFrame.toLowerCase();
		try {
			boolean del = FrameIO.DeleteFrame(toDelete);
			if (!del) {
				FrameGraphics.ErrorMessage("Error trying to delete "
						+ toDelete.getFrameName());
			} else {
				Frame current = DisplayIO.getCurrentFrame();
				for (Item i : current.getItems())
					if (i.getLink() != null
							&& i.getAbsoluteLink().toLowerCase().equals(
									deletedFrameNameLowercase)) {
						i.setLink(null);
					}

				FrameGraphics.Repaint();
				FrameGraphics.DisplayMessage(deletedFrame + " Deleted.");
			}
		} catch (IOException ioe) {
			FrameGraphics.ErrorMessage("Error trying to delete " + deletedFrame
					+ ":\n" + ioe.getMessage());
		}
	}

	/**
	 * Loads the Frame linked to by the given Item. The first Item on the Frame
	 * that is not the title or name is then placed on the cursor. If the given
	 * Item has no link, or no item is found then this is a no-op.
	 * 
	 * @param current
	 *            The Item that links to the Frame that the Item will be loaded
	 *            from.
	 */
	public static void GetItemFromChildFrame(Item current) {
		Item item = getFirstBodyItemOnChildFrame(current);
		// if no item was found
		if (item == null) {
			return;
		}

		// copy the item and switch
		item = item.copy();
		item.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		item.setParent(null);

		FrameMouseActions.pickup(item);
		FrameGraphics.Repaint();
	}

	/**
	 * Sets the given Item to have the Given Color. Color can be null (for
	 * default)
	 * 
	 * @param toChange
	 *            The Item to set the Color.
	 * @param toUse
	 *            The Color to give the Item.
	 */
	public static void SetCurrentItemBackgroundColor(Item toChange, Color toUse) {
		if (toChange == null)
			return;

		toChange.setBackgroundColor(toUse);
		FrameGraphics.Repaint();
	}

	/**
	 * Sets the given Item to have the Given Color. Color can be null (for
	 * default)
	 * 
	 * @param toChange
	 *            The Item to set the Color.
	 * @param toUse
	 *            The Color to give the Item.
	 */
	public static void SetCurrentItemColor(Item toChange, Color toUse) {
		if (toChange == null)
			return;

		toChange.setColor(toUse);
		FrameGraphics.Repaint();
	}

	/**
	 * Creates a new Text Object containing the Stats that are in the the
	 * JFrame's Title bar, The newly created Text Object is then attached to the
	 * cursor via FrameMouseActions.pickup(Item)
	 */
	public static void GetSessionStats() {
		String stats = SessionStats.getCurrentStats();

		Text text = DisplayIO.getCurrentFrame().createNewText();
		// We dont want the stats to wrap at all
		text.setMaxSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));

		text.setText(stats);

		text.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		text.setSize(STATS_FONT_SIZE);

		FrameMouseActions.pickup(text);

		FrameGraphics.Repaint();
	}

	/**
	 * Creates a new Text Object containing the contents of the current frames
	 * file.
	 */
	public static void GetCurrentFrameFile() {

		String fileContents = FrameIO.SaveFrame(DisplayIO.getCurrentFrame());

		Text text = DisplayIO.getCurrentFrame().createNewText();
		// We dont want the stats to wrap at all
		text.setMaxSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));

		text.setText(fileContents);

		text.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		text.setSize(FRAME_FILE_FONT_SIZE);

		FrameMouseActions.pickup(text);

		FrameGraphics.Repaint();
	}

	/**
	 * Creates a new Text Object containing the available fonts.
	 */
	public static void GetAvailableFontFamilyNames() {

		String[] availableFonts = GraphicsEnvironment
				.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
		StringBuilder fontsList = new StringBuilder();
		for (String s : availableFonts) {
			fontsList.append(s).append('\n');
		}

		Text text = DisplayIO.getCurrentFrame().createNewText();
		// We dont want the stats to wrap at all
		text.setMaxSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));
		text.setText(fontsList.toString());

		text.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		text.setSize(FRAME_FILE_FONT_SIZE);

		FrameMouseActions.pickup(text);

		FrameGraphics.Repaint();
	}

	/**
	 * Resets the statistics back to zero.
	 */
	public static void ResetStats() {
		StatsLogger.WriteStatsFile();
		SessionStats.resetStats();
	}

	/**
	 * Loads a frame with the given name and saves it as a JPEG image.
	 * 
	 * @param framename
	 *            The name of the Frame to save
	 */
	public static void JpegFrame(String framename) {
		ImageFrame(framename, "JPG");
	}

	/**
	 * Saves the current frame as a JPEG image. This is the same as calling
	 * JpegFrame(currentFrame.getFrameName())
	 */
	public static void JpegFrame() {
		ImageFrame(DisplayIO.getCurrentFrame().getFrameName(), "JPG");
	}

	/**
	 * Loads a frame with the given name and saves it as a PNG image.
	 * 
	 * @param framename
	 *            The name of the Frame to save
	 */
	public static void PNGFrame(String framename) {
		ImageFrame(framename, "PNG");
	}

	/**
	 * Saves the current frame as a PNG image. This is the same as calling
	 * PNGFrame(currentFrame.getFrameName())
	 */
	public static void PNGFrame() {
		ImageFrame(DisplayIO.getCurrentFrame().getFrameName(), "PNG");
	}

	/**
	 * Saves the Frame with the given Framename as an image of the given format.
	 * 
	 * @param framename
	 *            The name of the Frame to save as an image
	 * @param format
	 *            The Image format to use (i.e. "PNG", "BMP", etc)
	 */
	public static void ImageFrame(String framename, String format) {
		Frame loaded = FrameIO.LoadFrame(framename);
		String fileName = loaded.getExportFileName();

		// if the frame was loaded successfully
		if (loaded != null) {
			// check if the buffer needs to be redrawn
			// if (loaded.getBuffer() == null)
			FrameGraphics.UpdateBuffer(loaded, false);

			BufferedImage screen = loaded.getBuffer().getSnapshot();

			try {
				// set up the file for output
				File out = new File(FrameIO.EXPORTS_DIR + fileName + "."
						+ format.toLowerCase());
				if (!out.getParentFile().exists())
					out.mkdirs();

				ImageIO.write(screen, format, out);
				FrameGraphics.DisplayMessage("Frame successfully saved to "
						+ FrameIO.EXPORTS_DIR + out.getName());
			} catch (IOException e) {
				FrameGraphics.ErrorMessage(e.getMessage());
			}
			// if the frame was not loaded successfully, alert the user
		} else
			FrameGraphics.DisplayMessage("Frame '" + framename
					+ "' could not be found.");
	}

	/**
	 * Displays a message in the message box area.
	 * 
	 * @param message
	 *            the message to display
	 */
	public static void MessageLn(String message) {
		FrameGraphics.DisplayMessage(message);
	}

	public static void MessageLn2(String message, String message2) {
		FrameGraphics.DisplayMessage(message + " " + message2);
	}

	public static void CopyFile(String existingFile, String newFileName) {
		try {
			// TODO is there a built in method which will do this faster?

			FrameGraphics.DisplayMessage("Copying file " + existingFile
					+ " to " + newFileName + "...");
			FrameIO.copyFile(existingFile, newFileName);
			FrameGraphics.DisplayMessage("File copied successfully");
		} catch (FileNotFoundException e) {
			FrameGraphics.DisplayMessage("Error opening file: " + existingFile);
		} catch (Exception e) {
			FrameGraphics.DisplayMessage("File could not be copied");
		}
	}

	/**
	 * Loads the Frame linked to by the given Item. The first Item on the Frame
	 * that is not the title or name is then placed on the current frame. The
	 * item that was clicked on is placed on the frame it was linked to and the
	 * link is switched to the item from the child frame. If the given Item has
	 * no link, or no item is found then this is a no-op.
	 * 
	 * @param current
	 *            The Item that links to the Frame that the Item will be loaded
	 *            from.
	 */
	public static void SwapItemWithItemOnChildFrame(Item current) {
		Item item = getFirstBodyItemOnChildFrame(current);
		// if no item was found
		if (item == null) {
			return;
		}

		// swap the items parents
		Frame parentFrame = current.getParent();
		Frame childFrame = item.getParent();
		current.setParent(childFrame);
		item.setParent(parentFrame);

		// swap the items on the frames
		parentFrame.removeItem(current);
		childFrame.removeItem(item);
		parentFrame.addItem(item);
		childFrame.addItem(current);

		// swap the items links
		item.setAction(current.getAction());
		item.setLink(childFrame.getFrameName());
		current.setLink(parentFrame.getFrameName());
		// current.setLink(null);
		current.setAction(null);

		FrameGraphics.Repaint();
	}

	private static Item getFirstBodyItemOnChildFrame(Item current) {
		// the item must link to a frame
		if (current.getLink() == null) {
			FrameGraphics
					.DisplayMessage("Cannot get item from child - this item has no link");
			return null;
		}

		Frame child = FrameIO.LoadFrame(current.getAbsoluteLink());

		// if the frame could not be loaded
		if (child == null) {
			FrameGraphics.ErrorMessage("Could not load child frame.");
			return null;
		}

		// find the first non-title and non-name item
		List<Item> body = child.getItems();
		Item item = null;

		for (Item i : body)
			if (i != child.getTitle() && i != child.getFrameNameItem()
					&& !i.isAnnotation()) {
				item = i;
				break;
			}

		// if no item was found
		if (item == null) {
			FrameGraphics.DisplayMessage("No item found to copy");
			return null;
		}

		return item;
	}
}
