package org.expeditee.gui;

import java.awt.Color;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import org.expeditee.actions.Actions;
import org.expeditee.actions.Simple;
import org.expeditee.io.Logger;
import org.expeditee.items.Dot;
import org.expeditee.items.Item;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Text;
import org.expeditee.stats.SessionStats;

public class FrameKeyboardActions implements KeyListener {

	private static Text _toRemove = null;

	// these numbers correspond to the Function Key numbers (0 = Escape key)
	public static final int DROP_DOWN = 0;

	public static final int SIZE_UP = 1;

	public static final int SIZE_DOWN = 2;

	public static final int COLOR_CHANGE = 3;

	public static final int ANNOTATION_CHANGE = 4;

	public static final int DATE_ITEM = 5;

	public static final int NEW_FRAMESET = 6;

	public static final int FONT_STYLE_CHANGE = 7;

	public static final int FONT_FAMILY_CHANGE = 8;

	public static final int AUDIENCE_MODE = 9;

	public static final int XRAY_MODE = 10;

	// public static final int RUN_MENU = 11;

	public static final int FORCE_REPAINT = 12;

	public static final int MINIMUM_SIZE = 10;

	public synchronized void keyTyped(KeyEvent e) {
		if (Simple.isProgramRunning()) {
			Simple.KeyStroke(e.getKeyChar());
			return;
		}

		// ignore escape character and control characters
		if (e.getKeyChar() == KeyEvent.VK_ESCAPE || e.isControlDown()) {
			return;
		}
		// ingnore auto TDFC key sequence
		if (e.getKeyChar() == KeyEvent.VK_ENTER && e.isControlDown())
			return;

		e.consume();
		char ch = e.getKeyChar();

		if (e.isAltDown()) {
			if (FrameUtils.getCurrentItem() == null
					&& !Frame.itemAttachedToCursor()) {
				Text t = DisplayIO.getCurrentFrame().createNewText(ch + ": ");
				FrameMouseActions.pickup(t);
				FrameGraphics.Repaint();
			} else {
				// remove the link for alt+l
				if (ch == 'l') {
					FrameUtils.getCurrentItem().setLink(null);
					FrameGraphics.Repaint();
				}
			}
		} else {
			processChar(ch);
		}
		// FrameGraphics.Repaint();
	}

	public static void processChar(char ch) {

		Item on = FrameUtils.getCurrentItem();

		// permission check
		if (on != null && on.Permission < Item.PERMISSION_FULL) {
			FrameGraphics
					.DisplayMessage("Insufficient permission to edit this item");
			return;
		}

		if (_toRemove != null && on != _toRemove) {
			assert (_toRemove.getLength() == 0);
			// This line is to protect mistaken removal of items if there is a
			// bug...
			if (_toRemove.getLength() == 0)
				DisplayIO.getCurrentFrame().removeItem(_toRemove);
		}
		_toRemove = null;

		// ignore delete and backspace if in free space
		if (on == null
				&& (ch == KeyEvent.VK_BACK_SPACE || ch == KeyEvent.VK_TAB || ch == KeyEvent.VK_DELETE))
			return;

		SessionStats.TypedChar(ch);

		// check for dot's being replaced with text
		if (on != null && on instanceof Dot) {
			replaceDot((Dot) on, ch);
			return;
		}

		// only text can interact with keyboard events
		if (on != null && !(on instanceof Text))
			on = null;

		// DisplayIO.UpdateTitle();

		Text text = (Text) on;
		// if this text is empty but has not been removed (such as from
		// ESC-pushdown)
		if (text != null && text.isEmpty()
				&& (ch == KeyEvent.VK_BACK_SPACE || ch == KeyEvent.VK_DELETE)) {
			if (text.getLines().size() > 0)
				replaceText(text);
			else {
				DisplayIO.setCursor(Item.DEFAULT_CURSOR);
			}
			return;
		}

		// if the user is in free space, create a new text item
		if (on == null || !on.isHighlighted()) {
			SessionStats.CreatedItem();
			// DisplayIO.UpdateTitle();
			text = createText(ch);
			FrameUtils.LastEdited = text;
			DisplayIO.getCurrentFrame().addItem(text);
			DisplayIO.setTextCursor(text.getSize());
			return;
		}

		DisplayIO.setTextCursor(text.getSize());
		Point newMouse = text.insertChar(ch, DisplayIO.getMouseX(), DisplayIO
				.getMouseY());
		DisplayIO.setCursorPosition(newMouse.x, newMouse.y, false);

		// a change has occured to the Frame
		text.getParent().setChanged(true);

		// check that the Text item still exists (hasn't been deleted\backspaced
		// away)
		if (text.isEmpty()) {
			_toRemove = text;

			if (text.getAction() != null)
				text.setActionMark(true);
			else if (text.getLink() != null)
				text.setLinkMark(true);
			else if (text.getLines().size() > 0)
				replaceText(text);
			else {
				// DisplayIO.getCurrentFrame().removeItem(text);
				DisplayIO.setCursor(Item.DEFAULT_CURSOR);
			}
		}

		FrameUtils.LastEdited = text;
	}

	private static void replaceDot(Dot dot, char ch) {
		SessionStats.CreatedItem();
		Text text = createText(ch);
		FrameUtils.LastEdited = text;

		List<Line> lines = new LinkedList<Line>();
		lines.addAll(dot.getLines());
		for (Line line : lines)
			line.replaceEnd(dot, text);

		dot.removeAllLines();
		DisplayIO.getCurrentFrame().removeItem(dot);
		DisplayIO.getCurrentFrame().addItem(text);
		return;
	}

	// replaces the given text item with a dot
	private static void replaceText(Text text) {
		Dot dot = new Dot(text.getX(), text.getY(), text.getID());
		Item.DuplicateItem(text, dot);

		List<Line> lines = new LinkedList<Line>();
		lines.addAll(text.getLines());
		for (Line line : lines)
			line.replaceEnd(text, dot);
		text.removeAllLines();

		DisplayIO.getCurrentFrame().removeItem(text);
		DisplayIO.getCurrentFrame().addItem(dot);
		DisplayIO.setCursor(Item.DEFAULT_CURSOR);
	}

	/**
	 * Creates a new Text Item whose text contains the given character. This
	 * method also moves the mouse cursor to be pointing at the newly created
	 * Text Item ready to insert the next character.
	 * 
	 * @param start
	 *            The character to use as the initial text of this Item.
	 * @return The newly created Text Item
	 */
	private static Text createText(char start) {
		Text t = DisplayIO.getCurrentFrame().createBlankText("" + start);

		Point newMouse = t.insertChar(start, DisplayIO.getMouseX(), DisplayIO
				.getMouseY());
		DisplayIO.setCursorPosition(newMouse.x, newMouse.y, false);

		return t;
	}

	/**
	 * Creates a new Text Item with no text. The newly created Item is a copy of
	 * any ItemTemplate if one is present, and inherits all the attributes of
	 * the Template
	 * 
	 * @return The newly created Text Item
	 */
	private static Text createText() {
		return DisplayIO.getCurrentFrame().createNewText();
	}

	private void move(int direction) {
		Item ip = FrameUtils.getCurrentItem();

		if (ip == null) {
			navigateFrame(direction);
			return;
		}

		Item on = ip;
		if (on instanceof Text) {
			// When the user hits the left and right button with mouse
			// positions over the the frame name navigation occurs
			if (on.isFrameName()) {
				navigateFrame(direction);
				return;
			} else {
				FrameUtils.LastEdited = on;
				Point newMouse = ((Text) on).moveCursor(direction, DisplayIO
						.getMouseX(), DisplayIO.getMouseY());
				DisplayIO.setTextCursor(((Text) on).getSize());
				DisplayIO.setCursorPosition(newMouse, false);
			}
		}
	}

	private void navigateFrame(int direction) {
		Frame next = null;
		switch (direction) {
		case Text.RIGHT:
		case Text.PAGE_UP:
			next = FrameIO.LoadNext();
			break;
		case Text.LEFT:
		case Text.PAGE_DOWN:
			next = FrameIO.LoadPrevious();
			break;
		case Text.HOME:
			next = FrameIO.LoadZero();
			break;
		case Text.END:
			next = FrameIO.LoadLast();
			break;
		}
		FrameUtils.DisplayFrame(next, true);
	}

	/**
	 * Receives and processes any Function, Control, and Escape key presses
	 * 
	 * @param e
	 *            The KeyEvent received from the keyboard
	 */
	public void keyPressed(KeyEvent e) {
		SessionStats.AddFrameEvent("k" + KeyEvent.getKeyText(e.getKeyCode()));

		FrameUtils.ResponseTimer.restart();
		e.consume();

		if (Actions.isAgentRunning()) {
			if (e.getKeyCode() == KeyEvent.VK_ESCAPE)
				Actions.stopAgent();
			else
				Actions.interruptAgent();
			return;
		} else if (Simple.isProgramRunning()) {
			return;
		}

		if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
			functionKey(DROP_DOWN);
			SessionStats.Escape();
			return;
		} else if (e.getKeyCode() >= KeyEvent.VK_F1
				&& e.getKeyCode() <= KeyEvent.VK_F12) {
			functionKey(e.getKeyCode() - KeyEvent.VK_F1 + 1);
			return;
		} else if (e.isControlDown() && e.getKeyCode() != KeyEvent.VK_CONTROL)
			controlChar(e, KeyEvent.getKeyText(e.getKeyCode()));

		switch (e.getKeyCode()) {
		case KeyEvent.VK_LEFT:
			move(Text.LEFT);
			break;
		case KeyEvent.VK_RIGHT:
			move(Text.RIGHT);
			break;
		case KeyEvent.VK_PAGE_DOWN:
			navigateFrame(Text.PAGE_DOWN);
			break;
		case KeyEvent.VK_PAGE_UP:
			navigateFrame(Text.PAGE_UP);
			break;
		case KeyEvent.VK_UP:
			if (e.isControlDown() || e.isShiftDown()) {
				DisplayIO.getCurrentFrame();
			} else {
				move(Text.UP);
			}
			break;
		case KeyEvent.VK_DOWN:
			move(Text.DOWN);
			break;
		case KeyEvent.VK_END:
			move(Text.END);
			break;
		case KeyEvent.VK_HOME:
			move(Text.HOME);
			break;
		case KeyEvent.VK_ENTER:
			if (e.isControlDown()) {
				FrameMouseActions.leftButton(FrameUtils.getCurrentItem());
			}
			break;
		}
	}

	/**
	 * Currently ignored.
	 */
	public void keyReleased(KeyEvent e) {

	}

	/**
	 * 
	 * @param text
	 * @param yPos
	 * @return the new yPos for the next item
	 */
	private static int addTextItemToCurrentFrame(String text, int yPos) {
		if (text.trim().length() > 0) {
			Text textItem = createText();
			textItem.setText(text.substring(0, text.length() - 1));
			textItem.setY(yPos);
			DisplayIO.getCurrentFrame().addItem(textItem);
			return textItem.getY() + textItem.getBoundsHeight();
		}
		return yPos;
	}

	/**
	 * Processes all control character keystrokes. Currently Ctrl+C and Ctrl+V
	 * are copy and paste, all other keystrokes are ignored.
	 * 
	 * @param ch
	 *            The character being pressed along with the control key
	 */
	private static void controlChar(KeyEvent e, String ch) {
		// if this is a paste command
		if (ch.charAt(0) == KeyEvent.VK_V) {
			try {
				// read in the data from the clipboard
				String clip = ((String) Toolkit.getDefaultToolkit()
						.getSystemClipboard().getContents(null)
						.getTransferData(DataFlavor.stringFlavor));

				Item clicked = FrameUtils.getCurrentItem();

				if (clicked != null) {
					// check permissions
					if (clicked.Permission < Item.PERMISSION_FULL
							&& clicked.getParent() != null
							&& clicked.getParent().getFrameNameItem() != clicked) {
						FrameGraphics.DisplayMessage("Insufficient Permission");
						return;
					}

					Text text = createText();
					text.setText(clip);
					List<Item> clipboard = new ArrayList<Item>();
					clipboard.add(text);
					List<Item> left = FrameMouseActions.merge(clipboard,
							clicked);
					FrameMouseActions.anchor(left);
				} else {

					StringTokenizer st = new StringTokenizer(clip, "\n", true);

					String temp = "";
					int y = DisplayIO.getMouseY();

					// separate the clipboard content into items based on
					// blank lines
					while (st.hasMoreTokens()) {
						String s = st.nextToken();
						// if this is a blank line, then it is an item separator
						if (s.trim().length() == 0) {
							y = addTextItemToCurrentFrame(temp, y);
							temp = "";
						} else {
							temp += s + "\n";
							if (st.hasMoreTokens())
								st.nextToken();
						}
					}
					// the last item will not be finished by the above loop
					addTextItemToCurrentFrame(temp, y);
				}

				FrameGraphics.Repaint();
			} catch (Exception ex) {
			}
			return;

			// if this is a copy command
		} else if (ch.charAt(0) == KeyEvent.VK_C) {
			Item ip = FrameUtils.getCurrentItem();

			if (ip != null && ip.Permission < Item.PERMISSION_COPY) {
				FrameGraphics
						.DisplayMessage("Insufficient permission to copy that item");
				return;
			}

			Item on = null;
			if (ip != null)
				on = ip;

			if (on == null || !(on instanceof Text))
				return;

			String text = "";

			List<String> lines = ((Text) on).getText();
			for (String s : lines)
				text += s + "\n";
			// remove the last '\n'
			text = text.substring(0, text.length() - 1);

			// add the text of the item to the clipboard
			StringSelection selection = new StringSelection(text);
			Toolkit.getDefaultToolkit().getSystemClipboard().setContents(
					selection, null);

			FrameGraphics.DisplayMessage("Item copied to clipboard");
			return;
		} else if (ch.charAt(0) == KeyEvent.VK_D) {
			// perform a delete operation
			processChar((char) KeyEvent.VK_DELETE);
			FrameGraphics.Repaint();
		} else if (ch.charAt(0) == KeyEvent.VK_J) {

			Text text = getCurrentTextItem();
			if (text == null)
				return;

			if (text.getWidth() < 0)
				text.setWidth(text.getBoundsWidth() - Item.MARGIN_RIGHT
						- UserSettings.Gravity);

			text.setJustification(Item.JUSTIFICATION_LEFT);

			FrameGraphics.Repaint();
			return;
		} else if (ch.charAt(0) == KeyEvent.VK_S) {
			// split the current text item
			Text text = getCurrentTextItem();
			if (text == null)
				return;
			List<String> textLines = text.getText();
			if (textLines.size() <= 1)
				return;
			// remove all except the first line of text from the item being
			// split
			text.setText(textLines.get(0));
			int y = text.getY();
			for (int i = 1; i < textLines.size(); i++) {
				Text newText = text.copy();
				newText.setText(textLines.get(i));
				y += newText.getBoundsHeight();
				newText.setY(y);

				Frame current = DisplayIO.getCurrentFrame();
				// update the items ID to prevent conflicts with the current
				// frame
				newText.setID(current.getNextItemID());

				current.addItem(newText);
			}
			FrameGraphics.Repaint();

			return;
		}

		Logger.Log(Logger.USER, Logger.CONTROL_CHAR, "User pressing: Ctrl+"
				+ ch);
	}

	/**
	 * Gets the currently selected item if the user is allowed to modify it.
	 * 
	 * @return null if the currently selected item is not a Text item that the
	 *         user has permission to modify
	 */
	private static Text getCurrentTextItem() {
		Item ip = FrameUtils.getCurrentItem();

		if (ip != null && ip.Permission < Item.PERMISSION_FULL) {
			FrameGraphics
					.DisplayMessage("Insufficient permission to copy that item");
			return null;
		}

		Item on = null;
		if (ip != null)
			on = ip;

		if (on == null || !(on instanceof Text))
			return null;

		return (Text) on;
	}

	public static void functionKey(int key) {
		functionKey(key, 1);
	}

	/**
	 * Called when a Function key has been pressed, and performs the specific
	 * action based on the key.
	 */
	public static void functionKey(int key, int repeat) {
		// get whatever the user is pointing at
		Item on = FrameUtils.getCurrentItem();

		// check for enclosed mode
		if (on == null || on == null) {
			List<Item> enclosed = FrameUtils.getCurrentItems();

			if (enclosed != null && enclosed.size() > 0) {
				// ensure only one dot\line is present in the list
				List<Dot> dots = FrameUtils.getEnclosingDots();
				List<Item> connected = dots.get(0).getAllConnected();

				// only resize lines if they are not surrounding some non line
				// items
				boolean resizeLines = true;
				for (Item ip : enclosed) {
					if (ip != null) {
						if (!(ip instanceof Line) && !(ip instanceof Dot)) {
							resizeLines = false;
							break;
						}
					}
				}

				for (Item ip : enclosed) {
					if (ip != null && !(connected.indexOf(ip) > 0))
						switch (key) {
						case SIZE_UP:
							if (resizeLines
									|| (!(ip instanceof Line) && !(ip instanceof Dot)))
								SetSize(ip, repeat, false);
							break;
						case SIZE_DOWN:
							if (resizeLines
									|| (!(ip instanceof Line) && !(ip instanceof Dot)))
								SetSize(ip, -repeat, false);
							break;
						// case COLOR_CHANGE: SetFillColor(ip); break;
						case ANNOTATION_CHANGE:
							ToggleAnnotation(ip);
							break;
						case FONT_STYLE_CHANGE:
							ToggleFontStyle(ip);
							break;
						case FONT_FAMILY_CHANGE:
							ToggleFontFamily(ip);
							break;
						case DATE_ITEM:
							AddDate(ip);
							break;
						}
				}

				if (key == COLOR_CHANGE && connected.size() > 0) {
					for (Dot d : dots) {
						SetFillColor(d);
						break;
					}
				}
				return;
			}
		}

		switch (key) {
		case DROP_DOWN:
			Drop(on);
			break;
		case SIZE_UP:
			SetSize(on, repeat, true);
			break;
		case SIZE_DOWN:
			SetSize(on, -repeat, true);
			break;
		case COLOR_CHANGE:
			SetColor(on);
			break;
		case ANNOTATION_CHANGE:
			ToggleAnnotation(on);
			break;
		case FONT_STYLE_CHANGE:
			ToggleFontStyle(on);
			break;
		case FONT_FAMILY_CHANGE:
			ToggleFontFamily(on);
			break;
		case DATE_ITEM:
			AddDate(on);
			break;
		case NEW_FRAMESET:
			CreateFrameset(on);
			break;
		case XRAY_MODE:
			ToggleXRayMode(on);
			break;
		case AUDIENCE_MODE:
			ToggleAudience(on);
			break;
		// case RUN_MENU:
		// RunFirstMenuItem(on);
		// break;
		case FORCE_REPAINT:
			Repaint(on);
			break;
		}
	}

	public static final String DEFAULT_NEW_ITEM_TEXT = "";

	/**
	 * Performs the dropping action: If the cursor is in free space then: the
	 * cursor is repositioned below the last non-annotation text item. If the
	 * cursor is on an item, and has items attached then: the cusor is
	 * positioned below the pointed to item, and the items below are 'pushed
	 * down' to make room.
	 * 
	 * @param toDropFrom
	 *            The Item being pointed at by the mouse, may be null to
	 *            indicate the cursor is in free space.
	 */
	private static void Drop(Item ip) {
		String newItemText = DEFAULT_NEW_ITEM_TEXT;

		// if a line is being rubber-banded, this is a no-op
		if (Frame.rubberbandingLine())
			return; // No-op

		// drop requires no permissions
		Item toDropFrom = null;
		if (ip != null)
			toDropFrom = ip;

		// if the cursor is in free space then the drop will happen from the
		// last non annotation text item on the frame
		if (toDropFrom == null) {
			toDropFrom = DisplayIO.getCurrentFrame()
					.getLastNonAnnotationTextItem();
		}

		// if no item was found, return
		if (toDropFrom == null) {
			FrameGraphics.ErrorMessage("No item could be found to drop from");
			return;
		}

		if (!(toDropFrom instanceof Text)) {
			FrameGraphics.DisplayMessage("Only text items can be dropped from");
			return;
		}

		// Get the list of items that must be dropped
		List<Item> column = DisplayIO.getCurrentFrame().getColumn(toDropFrom);

		if (column == null) {
			FrameGraphics.ErrorMessage("No column found to align items to");
			return;
		}

		Text title = DisplayIO.getCurrentFrame().getTitle();

		// We wont do auto bulleting when dropping from titles
		if (toDropFrom != title) {
			newItemText = getAutoBullet(((Text) toDropFrom).getFirstLine());
		}

		Text dummyItem = null;

		if (Frame.textItemAttachedToCursor()) {
			dummyItem = (Text) Frame.getItemAttachedToCursor();
			String autoBullet = getAutoBullet(dummyItem.getTextNoList());

			if (autoBullet.length() > 0)
				newItemText = "";
			dummyItem.setText(newItemText + dummyItem.getTextNoList());
		}
		dummyItem = createText();
		if (Frame.textItemAttachedToCursor()) {
			Text t = (Text)Frame.getItemAttachedToCursor();
			dummyItem.setSize(t.getSize());
			for (int i = 0; i < t.getText().size(); i++) {
				newItemText += '\n';
			}
		}

		dummyItem.setText(newItemText);

		// If the only item on the frame is the title and the frame name just
		// drop a specified distance below the title
		if (column.size() == 0) {
			Text itemTemplate = DisplayIO.getCurrentFrame().getItemTemplate();
			int xPos = title.getX();
			int yPos = title.getY() + title.getBoundsHeight()
					+ itemTemplate.getBoundsHeight();
			dummyItem.setPosition(xPos, yPos);
			DisplayIO.setCursorPosition(xPos, yPos);
		} else {
			int yPos = column.get(0).getY() + 1;
			// Either position the new item below the title or just above
			// the first item below the title
			if (toDropFrom == title)
				yPos = Math
						.min(column.get(0).getY() - 1, title.getY()
								+ title.getBoundsHeight()
								+ dummyItem.getBoundsHeight());
			dummyItem.setPosition(column.get(0).getX(), yPos);
			column.add(dummyItem);
			FrameUtils.Align(column, false, 0);
			// Check if it will be outside the frame area
			if (dummyItem.getY() < 0
					|| dummyItem.getY() > FrameGraphics.getMaxFrameSize()
							.getHeight()) {
				// Check for the @more tag!
				Item i = ItemUtils.FindTag(DisplayIO.getCurrentFrame()
						.getItems(), "@More");
				if (i != null) {
					Frame firstFrame = DisplayIO.getCurrentFrame();
					boolean mouseMoved = FrameMouseActions.tdfc(i);
					Frame moreFrame = DisplayIO.getCurrentFrame();
					moreFrame.setTitle(firstFrame.getTitle().getTextNoList());
					// need to move the mouse to the top of the frame if there
					// wasnt an @start on it
					if (!mouseMoved) {
						Item moreTitle = moreFrame.getTitle();
						moreTitle.Permission = Item.PERMISSION_FULL;
						Drop(moreTitle);
					}
					// Add the bullet text to the item
					dummyItem.setPosition(DisplayIO.getMouseX(), DisplayIO
							.getMouseY());
				} else {
					FrameGraphics
							.WarningMessage("Can not create items outside the frame area");
					// ensures correct repainting when items don't move
					DisplayIO.setCursorPosition(DisplayIO.getMouseX(),
							DisplayIO.getMouseY());
					return;
				}
			}
			if (!Frame.textItemAttachedToCursor() && !dummyItem.isEmpty()) {
				DisplayIO.getCurrentFrame().addItem(dummyItem);
			}

			// Move the item to the cursor position
			if (Frame.itemAttachedToCursor()) {
				DisplayIO.setCursorPosition(dummyItem.getX(), dummyItem.getY());
				Item firstItem = Frame.getItemAttachedToCursor();
				int deltaX = firstItem.getX() - dummyItem.getX();
				int deltaY = firstItem.getY() - dummyItem.getY();

				for (Item i : Frame.FreeItems) {
					i.setPosition(i.getX() - deltaX, i.getY() - deltaY);
				}
			} else {
				DisplayIO
						.setCursorPosition(
								dummyItem.getEndParagraphPosition().x,
								dummyItem.getY());
			}
		}

		DisplayIO.resetCursorOffset();
		FrameGraphics.Repaint();
	}

	/**
	 * Gets the next letter sequence for a given string to be used in auto
	 * lettering.
	 * 
	 * @param s
	 *            a sequence of letters
	 * @return the next sequence of letters
	 */
	static private String nextLetterSequence(String s) {
		if (s.length() > 1)
			return s;

		if (s.equals("z"))
			return "a";

		return (char) ((int) s.charAt(0) + 1) + "";
	}

	public static String getBullet(String s) {
		return getBullet(s, false);
	}

	public static String getAutoBullet(String s) {
		return getBullet(s, true);
	}

	private static String getBullet(String s, boolean nextBullet) {
		String newItemText = DEFAULT_NEW_ITEM_TEXT;

		/*
		 * Item i = ItemUtils.FindTag(DisplayIO.getCurrentFrame().getItems(),
		 * "@NoAutoBullets"); if (i != null) return newItemText;
		 */

		// figure out the type of the text item
		// This allows us to do auto bulleting
		if (s != null && s.length() > 1) {
			// First check for text beginning with * @ # etc
			// These are simple auto bullets
			if (!Character.isLetterOrDigit(s.charAt(0))
					&& !Character.isSpaceChar(s.charAt(0))) {
				if (s.charAt(0) == '*' || s.charAt(0) == '+'
						|| s.charAt(0) == '-' || s.charAt(0) == '>'
						|| s.charAt(0) == '=' || s.charAt(0) == '@'
						|| s.charAt(0) == '~' || s.charAt(0) == '#') {
					int nonSpaceIndex = 1;
					while (nonSpaceIndex < s.length()
							&& s.charAt(nonSpaceIndex) == ' ') {
						nonSpaceIndex++;
					}
					// we must have a special char followed by >= 1 space
					if (nonSpaceIndex > 1)
						newItemText = s.substring(0, nonSpaceIndex);
				}
				// Auto numbering and lettering
			} else {
				if (Character.isDigit(s.charAt(0))) {
					newItemText = getAutoNumber(s, nextBullet);
					// Auto lettering
				} else if (Character.isLetter(s.charAt(0))) {
					newItemText = getAutoLetter(s, nextBullet);
				}
			}
		}
		return newItemText;
	}

	private static boolean isAutoNumberOrLetterChar(char c) {
		return c == ':' || c == '-' || c == '.' || c == ')' || c == '>';
	}

	/**
	 * Gets the string to be used to start the next auto numbered text item.
	 * 
	 * @param s
	 *            the previous text item
	 * @return the beginning of the next auto numbered text item
	 */
	private static String getAutoNumber(String s, boolean nextBullet) {
		String newItemText = DEFAULT_NEW_ITEM_TEXT;

		int nonDigitIndex = 1;
		while (Character.isDigit(s.charAt(nonDigitIndex))) {
			nonDigitIndex++;

			if (nonDigitIndex + 1 >= s.length())
				return DEFAULT_NEW_ITEM_TEXT;
		}

		if (isAutoNumberOrLetterChar(s.charAt(nonDigitIndex))) {

			// we must have a number followed one non letter
			// then one or more spaces
			int nonSpaceIndex = nonDigitIndex + 1;
			while (nonSpaceIndex < s.length() && s.charAt(nonSpaceIndex) == ' ') {
				nonSpaceIndex++;
			}

			if (nonSpaceIndex > nonDigitIndex + 1) {
				if (nextBullet)
					newItemText = (Integer.parseInt(s.substring(0,
							nonDigitIndex)) + 1)
							+ s.substring(nonDigitIndex, nonSpaceIndex);
				else
					newItemText = s.substring(0, nonSpaceIndex);
			}
		}
		return newItemText;
	}

	/**
	 * Gets the string to be used to start the next auto lettered text item.
	 * 
	 * @param s
	 *            the previous text items
	 * @return the initial text for the new text item
	 */
	private static String getAutoLetter(String s, boolean nextBullet) {
		String newItemText = DEFAULT_NEW_ITEM_TEXT;

		int nonLetterIndex = 1;

		if (isAutoNumberOrLetterChar(s.charAt(nonLetterIndex))) {

			// Now search for the next non space character
			int nonSpaceIndex = nonLetterIndex + 1;
			while (nonSpaceIndex < s.length() && s.charAt(nonSpaceIndex) == ' ') {
				nonSpaceIndex++;
			}

			// If there was a space then we have reached the end of our auto
			// text
			if (nonSpaceIndex > nonLetterIndex + 1) {
				if (nextBullet)
					newItemText = nextLetterSequence(s.substring(0,
							nonLetterIndex))
							+ s.substring(nonLetterIndex, nonSpaceIndex);
				else
					newItemText = s.substring(0, nonSpaceIndex);
			}
		}
		return newItemText;
	}

	/**
	 * Adjusts the size of the given Item, by the given amount. Note: The amount
	 * is relative and can be positive or negative.
	 * 
	 * @param toSet
	 *            The Item whose size is to be adjusted
	 * @param diff
	 *            The amount to adjust the Item's size by
	 * @param moveCursor
	 *            true if the cursor position should be automatically adjusted
	 *            with resizing
	 */
	private static void SetSize(Item ip, int diff, boolean moveCursor) {
		List<Item> toSize;
		// the mouse is only moved when the Item is on the frame, not free
		// boolean moveMouse = false;
		Item toSet = null;

		// if the user is not pointing to any item
		if (ip == null) {
			if (Frame.itemAttachedToCursor())
				toSize = new ArrayList<Item>(Frame.FreeItems);
			else {
				FrameGraphics
						.DisplayMessage("There are no Items selected on the Frame or on the Cursor");
				return;
			}
		} else {
			// check permissions
			if (ip.Permission < Item.PERMISSION_FULL) {
				FrameGraphics
						.DisplayMessage("Insufficient permission to change the size of that item");
				return;
			}
			toSet = ip;

			toSize = toSet.getConnected();
			// moveMouse = true;
		}

		if (toSet instanceof Line) {
			Line line = (Line) toSet;
			float current = Math.abs(line.getThickness());
			current = Math.max(current + diff, 1);
			line.setThickness(current);
			FrameGraphics.Repaint();
			return;
		}

		/*
		 * for(Item i : toSize){ if(i instanceof Dot || i instanceof Line){
		 * toSize.removeAll(i.getConnected()); toSize.add(i); } }
		 */

		int old_width = 0;
		int old_height = 0;

		if (toSet != null) {
			old_width = toSet.getBoundsWidth();
			old_height = toSet.getBoundsHeight();
		}

		// adjust the size of all the items
		for (Item i : toSize) {
			// Lines and dots use thickness, not size
			if (i instanceof Line) {
				Line line = (Line) i;
				float current = Math.abs(line.getThickness());
				current = Math.max(current + diff, 1);
				line.setThickness(current);
			} else if (i instanceof Dot) {
				Dot dot = (Dot) i;
				float current = Math.abs(dot.getThickness());
				current = Math.max(current + diff, 1);
				dot.setThickness(current);
			} else {
				int current = Math.abs(i.getSize());
				current = Math.max(current + diff, 1);
				if (current > MINIMUM_SIZE)
					i.setSize(current);
			}
		}

		// center the mouse cursor on the item
		if (moveCursor && toSet != null) {
			if (!toSet.contains(DisplayIO.getMouseX(), DisplayIO.getMouseY())) {
				int x = DisplayIO.getMouseX();
				int y = DisplayIO.getMouseY();

				if (!toSet.contains(x, toSet.getY()))
					x = x - (old_width - toSet.getBoundsWidth());

				// text grows 'up', pictures grow 'down'
				int direction = -1;
				if (toSet instanceof Text)
					direction = 1;

				if (!toSet.contains(toSet.getX(), y))
					y = y
							+ (direction * (old_height - toSet
									.getBoundsHeight()));

				DisplayIO.setCursorPosition(x, y);
			}

			// int x = toSet.getX() + toSet.getBoundsWidth() / 2;
			// int y = toSet.getY() + toSet.getBoundsHeight() / 2;;
			// DisplayIO.setCursorPosition(x, y);

		}

		if (toSet != null)
			toSet.getParent().setChanged(true);

		FrameGraphics.Repaint();
	}

	private static void SetFillColor(Item item) {
		if (item == null)
			return;

		Item toSet = item;// ;
		Color color = toSet.getFillColor();

		color = ColorUtils.getNextColor(color, Item.FILL_COLOR_WHEEL);

		// TODO what happens if the above statement returns null??

		toSet.setFillColor(color);
		toSet.getParent().setChanged(true);

		FrameGraphics.Repaint();
	}

	/**
	 * Sets the colour of the current Item based on its current colour. The
	 * colours proceed in the order stored in COLOR_WHEEL.
	 * 
	 * @param toSet
	 *            The Item whose colour is to be changed
	 */
	private static void SetColor(Item ip) {
		// first determine the next color
		Color color = null;

		if (ip == null) {
			if (Frame.itemAttachedToCursor()) {
				color = Frame.FreeItems.get(0).getColor();
			} else {
				return;
			}
			// change the background color if the user is pointing on the
			// frame name
		} else if (ip == DisplayIO.getCurrentFrame().getFrameNameItem()) {
			DisplayIO.getCurrentFrame().toggleBackgroundColor();
			return;
		} else {
			// check permissions
			if (ip.Permission < Item.PERMISSION_FULL) {
				FrameGraphics
						.DisplayMessage("Insufficient permission to change that item's color");
				return;
			}
			color = ip.getColor();
		}

		color = ColorUtils.getNextColor(color, Item.COLOR_WHEEL);

		// if we didnt find the color on the wheel
		if (color == null) {
			color = DisplayIO.getCurrentFrame().getItemTemplate().getColor();
		}

		if (Frame.itemAttachedToCursor()) {
			for (Item i : Frame.FreeItems)
				i.setColor(color);
		} else {
			Item toSet = ip;
			toSet.setColor(color);
			toSet.getParent().setChanged(true);

		}

		FrameGraphics.Repaint();
	}

	/**
	 * Toggles the given Item's annotation status on\off.
	 * 
	 * @param toToggle
	 *            The Item to toggle
	 */
	private static void ToggleAnnotation(Item toToggle) {
		if (toToggle == null) {
			FrameGraphics.DisplayMessage("There is no Item selected to toggle");
			return;
		}

		// check permissions
		if (toToggle.Permission < Item.PERMISSION_FULL) {
			FrameGraphics
					.DisplayMessage("Insufficient permission to toggle that item's annotation");
			return;
		}

		toToggle.setAnnotation(!toToggle.isAnnotation());
		toToggle.getParent().setChanged(true);
		FrameGraphics.Repaint();
	}

	/**
	 * Toggles the face style of a text item
	 * 
	 * @param toToggle
	 *            The Item to toggle
	 */
	private static void ToggleFontStyle(Item toToggle) {
		if (toToggle == null) {
			FrameGraphics.DisplayMessage("There is no Item selected to toggle");
			return;
		}

		// check permissions
		if (toToggle.Permission < Item.PERMISSION_FULL) {
			FrameGraphics
					.DisplayMessage("Insufficient permission to toggle that item's annotation");
			return;
		}

		if (toToggle instanceof Text) {
			Text text = (Text) toToggle;
			text.toggleFontStyle();

			text.getParent().setChanged(true);
			FrameGraphics.Repaint();
		}
	}

	/**
	 * Toggles the face style of a text item
	 * 
	 * @param toToggle
	 *            The Item to toggle
	 */
	private static void ToggleFontFamily(Item toToggle) {
		if (toToggle == null) {
			FrameGraphics.DisplayMessage("There is no Item selected to toggle");
			return;
		}

		// check permissions
		if (toToggle.Permission < Item.PERMISSION_FULL) {
			FrameGraphics
					.DisplayMessage("Insufficient permission to toggle that item's annotation");
			return;
		}

		if (toToggle instanceof Text) {
			Text text = (Text) toToggle;
			text.toggleFontFamily();

			text.getParent().setChanged(true);
			FrameGraphics.Repaint();
		}
	}

	/**
	 * If the given Item is null, then a new Text item is created with the
	 * current date If the given Item is not null, then the current date is
	 * prepended to the Item's text
	 * 
	 * @param toAdd
	 *            The Item to prepend the date to, or null
	 */
	private static void AddDate(Item toAdd) {
		String date1 = Logger.EasyDateFormat("ddMMMyyyy[HH:mm]");
		String date2 = Logger.EasyDateFormat("ddMMMyyyy");
		final String leftSeparator = " :";
		final String rightSeparator = ": ";
		String dateToAdd = date1 + rightSeparator;
		boolean prepend = false;
		boolean append = false;

		// if the user is pointing at an item, add the date where ever the
		// cursor is pointing
		if (toAdd != null) {
			if (toAdd instanceof Text) {
				// permission check
				if (toAdd.Permission < Item.PERMISSION_FULL) {
					FrameGraphics
							.DisplayMessage("Insufficicent permission to add the date to that item");
					return;
				}

				Text textItem = (Text) toAdd;

				String text = textItem.getTextNoList();

				// check if the default date has already been put on this item
				if (text.startsWith(date1 + rightSeparator)) {
					textItem.removeText(date1 + rightSeparator);
					dateToAdd = date2 + rightSeparator;
					prepend = true;
				} else if (text.startsWith(date2 + rightSeparator)) {
					textItem.removeText(date2 + rightSeparator);
					dateToAdd = leftSeparator + date2;
					append = true;
				} else if (text.endsWith(leftSeparator + date2)) {
					textItem.removeEndText(leftSeparator + date2);
					append = true;
					dateToAdd = leftSeparator + date1;
				} else if (text.endsWith(leftSeparator + date1)) {
					textItem.removeEndText(leftSeparator + date1);
					if (textItem.getLength() > 0) {
						dateToAdd = "";
						prepend = true;
					} else {
						// use the default date format
						prepend = true;
					}
				}

				if (prepend) {
					// add the date to the text item
					textItem.prependText(dateToAdd);
					if (dateToAdd.length() == textItem.getLength())
						DisplayIO.setCursorPosition(textItem
								.getEndParagraphPosition());
				} else if (append) {
					textItem.appendText(dateToAdd);
					if (dateToAdd.length() == textItem.getLength())
						DisplayIO.setCursorPosition(textItem.getPosition());
				} else {
					for (int i = 0; i < date1.length(); i++) {
						processChar(date1.charAt(i));
					}
				}

				textItem.getParent().setChanged(true);
				FrameGraphics.Repaint();
			} else {
				FrameGraphics
						.DisplayMessage("Only text items can have the date prepended to them");
			}
			// otherwise, create a new text item
		} else {
			Text newText = createText();
			newText.setText(dateToAdd);
			DisplayIO.getCurrentFrame().addItem(newText);
			DisplayIO.getCurrentFrame().setChanged(true);
			FrameGraphics.Repaint();

			DisplayIO.setCursorPosition(newText.getEndParagraphPosition());
		}

	}

	/**
	 * Creates a new Frameset with the name given by the Item
	 * 
	 * @param name
	 */
	private static void CreateFrameset(Item ip) {
		if (ip == null || !(ip instanceof Text)) {
			FrameGraphics
					.DisplayMessage("There is no selected item to use for the Frameset name");
			return;
		}

		// check permissions
		if (ip.Permission < Item.PERMISSION_FULL) {
			FrameGraphics
					.DisplayMessage("Insufficient permission to create a frameset from this item");
			return;
		}

		Text text = (Text) ip;
		try {
			// create the new frameset
			Frame linkTo = FrameIO.CreateNewFrameset(text.getFirstLine());
			DisplayIO.setCursor(Item.DEFAULT_CURSOR);
			text.setLink(linkTo.getFrameName());
			text.getParent().setChanged(true);
			FrameUtils.DisplayFrame(linkTo, true);
			linkTo.moveMouseToDefaultLocation();
			// this needs to be done if the user doesnt move the mouse before
			// doing Tdfc while the cursor is set to the text cursor
			DisplayIO.setCursor(Item.DEFAULT_CURSOR);
		} catch (Exception e) {
			FrameGraphics.ErrorMessage(e.getMessage());
		}
	}

	/**
	 * Toggles Audience mode on\off
	 * 
	 * @param ignored
	 *            This Item is currently ignored
	 */
	private static void ToggleAudience(Item ignored) {
		FrameGraphics.ToggleAudienceMode();
	}

	private static void ToggleXRayMode(Item ignored) {
		FrameGraphics.ToggleXRayMode();
	}

	/**
	 * Runs the first Item on the first menu
	 * 
	 * @param ignored
	 *            This Item is currently ignored
	 */
	/*
	 * private static void RunFirstMenuItem(Item ignored) {
	 * DisplayIO.activateMenuItem(0, 0); }
	 */

	/**
	 * Forces a re-parse and repaint of the current Frame.
	 * 
	 * @param ignored
	 *            This Item is currently ignored
	 */
	private static void Repaint(Item ignored) {
		FrameUtils.Parse(DisplayIO.getCurrentFrame());
		FrameGraphics.ForceRepaint();
	}
}
